"""Extremely simple demo functions."""


import typing as t
import logging


log = logging.getLogger(__name__)


def process(data: t.Dict[bytes, bytes]) -> None:
    """Prints the incoming data to stdout."""
    log.info("GOT DATA: %s", data)



"""Topix - Functional Microframework for Redis Streams

Topix is a microframework to create event streams in Redis, using
a functional programming style in Python.
"""


import logging

from .emitter import emit
from .stream import stream_into

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Streams module - functions that interact with Redis streams."""


import logging

import typing as t
import multiprocessing as mp

from functools import partial as p
from contextlib import contextmanager

import redis

from topix.utils import (
    concurrently,
    lazy_unpack,
    split_into,
    consume,
    generator,
    connection,
    guid,
)


T = t.TypeVar("T")


# Mapping stored in Redis streams.
StreamData = t.Dict[bytes, bytes]
StreamEntry = t.Tuple[bytes, StreamData]


log = logging.getLogger(__name__)


def _trim(x: t.List[t.Tuple[t.Any, T]]) -> T:
    """Removes extranneous data from Redis xreadgroup."""
    return x[0][1]


def _setup(stream: str, group: str) -> None:
    """Setup a stream with a consumer group on Redis."""
    try:
        connection().xgroup_create(stream, group, mkstream=True)  # type: ignore
        log.info("Created group=%s in stream=%s", group, stream)
    except redis.exceptions.ResponseError:
        log.info("Already created group=%s in stream=%s", group, stream)


def _teardown(stream: str, group: str, consumer: str) -> None:
    """Teardown a consumer in the group and stream."""
    log.warning(
        "Removing consumer=%s from group=%s on stream=%s", consumer, group, stream
    )
    connection().xgroup_delconsumer(stream, group, consumer)  # type: ignore


@contextmanager
def lifecycle(stream: str, group: str, consumer: str) -> t.Iterator[None]:
    """Redis stream lifecycle."""
    _setup(stream, group)
    try:
        yield
    finally:
        _teardown(stream, group, consumer)


def stream_into(
    fn: t.Callable[[StreamData], None],
    stream: str,
    group: str,
    consumer: t.Optional[str] = None,
    concurrency: int = mp.cpu_count(),
) -> None:
    """Feed entries from a Redis stream into a callable fn()

    Imagining a Redis stream as an infinite iterator, stream_into() applies
    the function fn to each element of the stream.

    Under the hood, stream_into() makes use of Python's concurrency primitives
    to allow for better performance of IO-heavy tasks on a consumer.

    Arguments:
        fn: Callable to map over elements of the stream.
        stream: Name of the redis stream.
        group: Name of the consumer group.
    Keyword Arguments:
        consumer: ID of the consumer to use. By default, a random UUIDv4
        concurrency: Number of threads to use. By default, the number of CPUs.
    Returns:
        None
    """
    consumer = consumer or guid()
    client = connection()

    # Functions to interact with Redis
    read = p(client.xreadgroup, group, consumer, {stream: ">"}, block=0)
    xack = p(client.xack, stream, group)

    # Transform nested redis-py xreadgroup response into an infinite lazy generator.
    source: t.Iterator[StreamEntry] = lazy_unpack(map(_trim, generator(read)))
    mapped = map(split_into(xack, fn), source)

    with lifecycle(stream, group, consumer):
        concurrently(mapped, threads=concurrency)



"""Command line script."""


import typing as t

import argparse
import logging
import json
import sys

from importlib import import_module


from topix.emitter import emit, CastType
from topix.stream import stream_into


# I am not even going to pretend this is typesafe...
def import_from(path: str) -> t.Any:
    """Import something from a python path."""
    module, item = path.split(":")
    return getattr(import_module(module), item)


def emitter(stream: str, mapping: t.Dict[CastType, CastType], **_: t.Any) -> None:
    """Emit a mapping to a stream from the command line."""
    emit(stream, mapping)


def consumer(
    stream: str,
    group: str,
    function: str,
    consumer: t.Optional[str] = None,
    concurrency: t.Optional[int] = None,
    **_: t.Any,
) -> None:
    """Consume events from the stream with a function."""
    fn = import_from(function)
    stream_into(fn, stream, group, consumer=consumer)


def _create_parsers() -> argparse.ArgumentParser:
    """Create a parser for the program."""
    parser = argparse.ArgumentParser(description=main.__doc__, prog="topix")

    # Global pieces, like logging configuration.
    parser.add_argument(
        "--log-format", default="%(levelname)s | %(name)s | %(asctime)s | %(message)s"
    )
    parser.add_argument("--log-level", default="INFO")

    subparsers = parser.add_subparsers()

    emit_parser = subparsers.add_parser("emitter", help=emitter.__doc__)
    emit_parser.add_argument("stream", type=str, help="Name of the stream to emit to.")
    emit_parser.add_argument(
        "mapping", type=json.loads, help="JSON-compatible string to emit."
    )
    emit_parser.set_defaults(func=emitter)

    consume_parser = subparsers.add_parser("consumer", help=consumer.__doc__)
    consume_parser.add_argument(
        "stream", type=str, help="Name of the stream to consume events from"
    )
    consume_parser.add_argument(
        "group", help="Name of the Redis consumer group to participate in"
    )
    consume_parser.add_argument(
        "function", type=str, help="Import path to the function (ex. 'app.myfunction')",
    )
    consume_parser.add_argument(
        "--consumer", type=str, help="Unique identifier for this consumer."
    )
    consume_parser.add_argument(
        "--concurrency", type=int, help="Number of threads to run the function with."
    )
    consume_parser.set_defaults(func=consumer)

    return parser


def main(raw_arguments: t.List[str]) -> None:
    """Topix CLI - Control Topix processes and streams.
    
    The command line allows the starting of "worker" processes,
    along with other miscellaneous tasks.
    """
    parser = _create_parsers()
    args = parser.parse_args(raw_arguments)
    logging.basicConfig(format=args.log_format, level=args.log_level)
    args.func(**vars(args))


if __name__ == "__main__":
    main(sys.argv[1:])



"""Emit messages to the stream."""


import typing as t

from topix.utils import connection


CastType = t.Union[str, int, bytes]


def emit(
    stream: str, data: t.Dict[CastType, CastType], limit: t.Optional[int] = None
) -> None:
    """Emit messages to the Redis stream.

    While `emit` accepts messages with values that are strings, integers,
    or bytes, the types are automatically cast to `bytes` before submission
    to the stream, and are similarly `bytes` when being read.

    To send more complex objects over Redis, consider using a serialization
    format like Pickle or JSON, and storing the bytes that way.

    Arguments:
        stream: Name of the stream to emit to.
        data: Data to send to the stream.
    Returns:
        None
    """
    client = connection()
    if limit is not None:
        client.xadd(stream, data, maxlen=limit, approximate=True)  # type: ignore
    else:
        client.xadd(stream, data)  # type: ignore



"""Utilities used for Topix."""


import os

import typing as t
import multiprocessing.dummy as mp

from functools import lru_cache
from uuid import uuid4

from collections import deque
from uuid import uuid4

import redis


# There's a few higher order functions in here- we need some
# type variables.
T = t.TypeVar("T")
A = t.TypeVar("A")
B = t.TypeVar("B")


@lru_cache(maxsize=1)
def connection() -> redis.Redis:
    """Creates a singleton Redis client."""
    # Allow configuration of what env var to use.
    url = os.getenv("TOPIX_REDIS_URL", "redis://localhost:6379/0")
    return redis.Redis.from_url(url)


def generator(fn: t.Callable[[], T]) -> t.Iterator[T]:
    """Turn a callable function into an infinite generator."""
    while True:
        yield fn()


def consume(iterable: t.Iterator[T]) -> None:
    """Completely consume an iterable. Usually used with map()"""
    deque(iterable, maxlen=0)


def lazy_unpack(iterable: t.Iterator[t.Iterator[T]]) -> t.Iterator[T]:
    """Convert an iterator of iterators into an iterator of items.

    iter([[x, y, z], [a, b, c]]) -> iter([x, y z, a, b, c])
    """
    for member in iterable:
        yield from member


def concurrently(iterable: t.Iterator[T], threads: int = 4) -> None:
    """Concurrently iterate through an iterable.

    Arguments:
        iterable: Iterator to move through.
    Keyword Arguments:
        threads: Number of threads to use.
    Returns:
        None
    """
    with mp.Pool(threads) as p:
        p.map(consume, iterable)


def split_into(
    x: t.Callable[[A], T], y: t.Callable[[B], None]
) -> t.Callable[[t.Tuple[A, B]], T]:
    """Create a function to unpack a tuple and use one as a side effect."""

    def z(val: t.Tuple[A, B]) -> T:
        """Send val[0] to x, val[0] to y, and return x(val[0])."""
        a, b = val
        result = x(a)
        y(b)
        return result

    return z


def guid() -> str:
    """Create a random UUID4 string."""
    return str(uuid4())
